#include "ape.h"
#include "cnv_ape.h"

APE::APE () {
    buffer = 0;
    buffer_size = 0;
    pAPEDecompress = 0;
    needseek = -1;
}

APE::~APE () {
    if ( buffer ) free ( buffer );
    if ( pAPEDecompress ) delete pAPEDecompress;
}

int APE::init ( MediaInfo *infos ) {
    if ( pAPEDecompress ) return 1;

    m_reader = infos->getReader();

    pAPEDecompress = CreateIAPEDecompressEx ( this );
    if ( !pAPEDecompress ) return 0;
    needseek = -1;

    return 1;
}

int APE::processData ( MediaInfo *infos, ChunkList *chunk_list, bool *killswitch ) {
    if ( !init (infos) ) return 0;

    if ( needseek >= 0 ) {
        double secs = (double)needseek * 1.e-3;
        needseek = -1;
		if ( pAPEDecompress->Seek ((int)(secs * (double)pAPEDecompress->GetInfo(APE_INFO_SAMPLE_RATE))) != 0 ) return 0;
    }

    int nBlocksRetrieved = 0;
    int bufsize = pAPEDecompress->GetInfo ( APE_INFO_BLOCK_ALIGN ) * DELTA_SAMPLES;
    if ( bufsize > buffer_size ) {
        void *backup = (void *)buffer;
        buffer_size = bufsize;
        buffer = (char *)realloc ( buffer, buffer_size );
        if ( !buffer ) {
            if ( backup ) free ( backup );
            return 0;
        }
    }
    char *ptr = (char *)buffer;
    int nRetVal = pAPEDecompress->GetData ( ptr, DELTA_SAMPLES, &nBlocksRetrieved );
    if ( nRetVal != ERROR_SUCCESS || nBlocksRetrieved <= 0 ) return 0;

    ChunkInfosI *ci = new ChunkInfosI();
    ci->addInfo ( "srate", pAPEDecompress->GetInfo (APE_INFO_SAMPLE_RATE) );
    ci->addInfo ( "bps", pAPEDecompress->GetInfo (APE_INFO_BITS_PER_SAMPLE) );
    ci->addInfo ( "nch", pAPEDecompress->GetInfo (APE_INFO_CHANNELS) );
    chunk_list->setChunk ( "PCM", ptr, nBlocksRetrieved * pAPEDecompress->GetInfo (APE_INFO_BLOCK_ALIGN), ci );

    return 1;
}

int APE::getInfos ( MediaInfo *infos ) {
    if ( !init (infos) ) return 0;

    if ( ReadTags (infos) == 0 ) {
        infos->setTitle ( Std::filename (infos->getFilename()) );
    }

    char temp[256];
    double length = (double)pAPEDecompress->GetInfo ( APE_DECOMPRESS_LENGTH_MS ) / 1000.;
    int channels = pAPEDecompress->GetInfo ( APE_INFO_CHANNELS );
    int srate = pAPEDecompress->GetInfo ( APE_INFO_SAMPLE_RATE );
    int bitrate = ( m_reader->canSeek() && length != 0 )  ?  bitrate = (int)((double)m_reader->getLength() * 8. * 1.e-3 / length)  :  pAPEDecompress->GetInfo ( APE_INFO_AVERAGE_BITRATE );
    int bps = pAPEDecompress->GetInfo ( APE_INFO_BITS_PER_SAMPLE );
    int compression = pAPEDecompress->GetInfo ( APE_INFO_COMPRESSION_LEVEL );
    int flags = pAPEDecompress->GetInfo ( APE_INFO_FORMAT_FLAGS );
    int version = pAPEDecompress->GetInfo ( APE_INFO_FILE_VERSION );

    infos->setLength ( (int)(length * 1000.) );

    if ( version%10 == 0 ) {
        SPRINTF ( temp, "%i.%i", version/1000, (version/10)%100 );
    } else {
        SPRINTF ( temp, "%i.%i", version/1000, version%1000 );
    }
    infos->setData ( "version", temp, MediaInfo::INFO_VISIBLE );

    switch ( compression ) {
    case COMPRESSION_LEVEL_FAST:
        STRCPY ( temp, "Fast" );
        break;
    case COMPRESSION_LEVEL_NORMAL:
        STRCPY ( temp, "Normal" );
        break;
    case COMPRESSION_LEVEL_HIGH:
        STRCPY ( temp, "High" );
        break;
    case COMPRESSION_LEVEL_EXTRA_HIGH:
        STRCPY ( temp, "Extra High" );
        break;
    case COMPRESSION_LEVEL_INSANE:
        STRCPY ( temp, "Insane" );
        break;
    default:
        STRCPY ( temp, "unknown" );
        break;
    }
    infos->setData ( "compression level", temp, MediaInfo::INFO_VISIBLE );

    infos->setDataInt ( "bitrate", bitrate, MediaInfo::INFO_VISIBLE );

    infos->setDataInt ( "format flags", flags, MediaInfo::INFO_VISIBLE );

    infos->setDataInt ( "sample rate", srate, MediaInfo::INFO_VISIBLE );

    infos->setDataInt ( "channels", channels, MediaInfo::INFO_VISIBLE );

    infos->setDataInt ( "bits per sample", bps, MediaInfo::INFO_VISIBLE );

    // update status display
    char str_status[256];
    char *p = str_status;

    p += SPRINTF ( p, "%ikbps ", bitrate );

    if ( channels > 0 ) {
        if ( channels == 1 )
            p += SPRINTF ( p, "mono " );
        else if ( channels == 2 )
            p += SPRINTF ( p, "stereo " );
        else
            p += SPRINTF ( p, "%i channels ", channels );
    }

    if ( srate > 0 ) {
        if ( srate < 1000 )
            p += SPRINTF ( p, "%iHz ", srate );
        else if ( srate % 1000 == 0 )
            p += SPRINTF ( p, "%ikHz ", srate / 1000 );
        else
            p += SPRINTF ( p, "%i.%ikHz ", srate / 1000, (srate % 1000) / 100 );
    }

    infos->setInfo ( str_status );

    return 1;
}

int APE::corecb_onSeeked ( int newpos ) {
    if ( !pAPEDecompress ) return 0;
    needseek = newpos;
    return 1;
}

int APE::canConvertFrom ( svc_fileReader *reader, const char *name, const char *chunktype ) {
    if ( chunktype || !name || !reader ) return 0;

    if ( reader->canSeek() ) {
        char header[3];
        int tagsize = SkipID3v2Tag ( reader );
        if ( tagsize < 0 ) return 0;
        int r = reader->read ( header, 3 );
        reader->seek ( 0 );

        if ( r < 3 || MEMCMP (header, "MAC", 3) ) return 0;

        return 1;
    }

    return !STRICMP ( Std::extension (name), "ape" ) || !STRICMP ( Std::extension (name), "mac" );
}
